@@ -41,7 +41,7 @@
 
 */
 
-#ifdef WIN32
+#ifdef _WIN32
 #include <winsock.h>
 #else
 #include <sys/param.h>
@@ -67,6 +67,9 @@ long fli_camera_parport_open(flidev_t dev)
 
   cam = DEVICE->device_data;
 
+	cam->removebias = 1;
+	cam->biasoffset = 200;
+
   /* Set timeout values */
   cam->readto = 1000;
   cam->writeto = 1000;
@@ -599,7 +602,7 @@ long fli_camera_parport_expose_frame(flidev_t dev)
   }
 
   debug(FLIDEBUG_INFO, "Exposing.");
-  buf = htons((unsigned short) C_SHUTTER((cam->frametype == FLI_FRAME_TYPE_DARK)?0:1,
+  buf = htons((unsigned short) C_SHUTTER((cam->frametype & FLI_FRAME_TYPE_DARK)?0:1,
 			cam->expmul));
   IO(dev, &buf, &wlen, &rlen);
 
@@ -762,8 +765,9 @@ long fli_camera_parport_write_ioport(flidev_t dev, long ioportset)
 {
   long rlen, wlen;
   unsigned short buf;
+	unsigned short _ioportset = (unsigned short) ioportset;
 
-  correctioportdatawrite(dev, (unsigned short *) &ioportset);
+  correctioportdatawrite(dev, &_ioportset);
   buf = htons((unsigned short) (0x7100 | (ioportset & 0x00ff)));
 
   rlen = 2; wlen = 2;
@@ -776,8 +780,9 @@ long fli_camera_parport_configure_ioport(flidev_t dev, long ioportset)
 {
   long rlen, wlen;
   unsigned short buf;
+	unsigned short _ioportset = (unsigned short) ioportset;
 
-  correctioportdatawrite(dev, (unsigned short *) &ioportset);
+  correctioportdatawrite(dev, &_ioportset);
   buf = htons((unsigned short) (0x7000 | (ioportset & 0x00ff)));
 
   rlen = 2; wlen = 2;
