@@ -65,7 +65,6 @@ const fliccdinfo_t knowndev[] = {
   {9,  "TK-1024",       {{0, 0}, {1124, 1024}}, {{50, 0},  {1074, 1024}}, 1.0, 24.0, 24.0},
   {10, "TK-512",        {{0, 0}, {563,  512}},  {{51, 0},  {563,   512}}, 1.0, 24.0, 24.0},
   {11, "SI-003A",       {{0, 0}, {1056, 1024}}, {{16, 0},  {1040, 1024}}, 1.0, 24.0, 24.0},
-//  {11, "SI-003A",       {{0, 0}, {1120, 1024}}, {{0, 0},  {1120, 1024}}, 1.0, 24.0, 24.0},
   {12, "KAF-6300",      {{0, 0}, {3100, 2056}}, {{16, 4},  {3088, 2052}}, 1.0,  9.0,  9.0},
   {13, "KAF-3200",      {{0, 0}, {2267, 1510}}, {{46,34},  {2230, 1506}}, 1.0,  6.8,  6.8},
   {14, "SI424A",        {{0, 0}, {2088, 2049}}, {{20, 0},  {2068, 2049}}, 1.0,  6.8,  6.8},
@@ -139,6 +138,12 @@ long fli_camera_close(flidev_t dev)
     cam->gbuf = NULL;
   }
 
+	 if (cam->ibuf != NULL)
+  {
+    xfree(cam->ibuf);
+    cam->ibuf = NULL;
+  }
+
   if (DEVICE->devinfo.model != NULL)
   {
     xfree(DEVICE->devinfo.model);
@@ -257,9 +262,9 @@ long fli_camera_command(flidev_t dev, int cmd, int argc, ...)
 				r = -EINVAL;
 			else
 			{
-				long exptime;
+				unsigned long exptime;
 
-				exptime = *va_arg(ap, long *);
+				exptime = *va_arg(ap, unsigned long *);
 
 				switch (DEVICE->domain)
 				{
@@ -436,6 +441,106 @@ long fli_camera_command(flidev_t dev, int cmd, int argc, ...)
 			}
 			break;
 
+		case FLI_SET_FAN_SPEED:
+			if (argc != 1)
+				r = -EINVAL;
+			else
+			{
+				long fan_speed;
+
+				fan_speed = *va_arg(ap, long *);
+
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_set_fan_speed(dev, fan_speed);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
+		case FLI_SET_VERTICAL_TABLE_ENTRY:
+			if (argc != 4)
+				r = -EINVAL;
+			else
+			{
+				long index;
+				long height;
+				long bin;
+				long mode;
+
+				index = *va_arg(ap, long *);
+				height = *va_arg(ap, long *);
+				bin = *va_arg(ap, long *);
+				mode = *va_arg(ap, long *);
+
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_set_vertical_table_entry(dev, index, height, bin, mode);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
+		case FLI_ENABLE_VERTICAL_TABLE:
+			if (argc != 3)
+				r = -EINVAL;
+			else
+			{
+				long width;
+				long offset;
+				long flags;
+
+				width = *va_arg(ap, long *);
+				offset = *va_arg(ap, long *);
+				flags = *va_arg(ap, long *);
+
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_enable_vertical_table(dev, width, offset, flags);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
+		case FLI_GET_VERTICAL_TABLE_ENTRY:
+			if (argc != 4)
+				r = -EINVAL;
+			else
+			{
+				long index;
+				long *height;
+				long *bin;
+				long *mode;
+
+				index = *va_arg(ap, long *);
+				height = va_arg(ap, long *);
+				bin = va_arg(ap, long *);
+				mode = va_arg(ap, long *);
+
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_get_vertical_table_entry(dev, index, height, bin, mode);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
 		case FLI_SET_TEMPERATURE:
 			if (argc != 1)
 				r = -EINVAL;
@@ -513,6 +618,29 @@ long fli_camera_command(flidev_t dev, int cmd, int argc, ...)
 			}
 			break;
 
+		case FLI_SET_TDI:
+			if (argc != 2)
+				r = -EINVAL;
+			else
+			{
+				flitdirate_t rate;
+				flitdiflags_t flags;
+
+				rate = *va_arg(ap, flitdirate_t *);
+				flags = *va_arg(ap, flitdiflags_t *);
+
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_set_tdi(dev, rate, flags);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
 		case FLI_GRAB_ROW:
 			if (argc != 2)
 				r = -EINVAL;
@@ -561,6 +689,64 @@ long fli_camera_command(flidev_t dev, int cmd, int argc, ...)
 			}
 			break;
 
+		case FLI_START_VIDEO_MODE:
+			if (argc != 0)
+				r = -EINVAL;
+			else
+			{
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_start_video_mode(dev);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
+		case FLI_STOP_VIDEO_MODE:
+			if (argc != 0)
+				r = -EINVAL;
+			else
+			{
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_stop_video_mode(dev);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
+		case FLI_GRAB_VIDEO_FRAME:
+			if (argc != 2)
+				r = -EINVAL;
+			else
+			{
+				void *buf;
+				size_t size;
+
+				buf = va_arg(ap, void *);
+				size = *va_arg(ap, size_t *);
+
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_grab_video_frame(dev, buf, size);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
+
 		case FLI_FLUSH_ROWS:
 			if (argc != 2)
 				r = -EINVAL;
@@ -712,6 +898,44 @@ long fli_camera_command(flidev_t dev, int cmd, int argc, ...)
 			}
 			break;
 
+		case FLI_GET_READOUT_DIMENSIONS:
+			if (argc != 6)
+				r = -EINVAL;
+			else
+			{
+				long *width, *hoffset, *hbin, *height, *voffset, *vbin;
+				flicamdata_t *cam = DEVICE->device_data;
+
+				width = va_arg(ap, long *);
+				hoffset = va_arg(ap, long *);
+				hbin = va_arg(ap, long *);
+				height = va_arg(ap, long *);
+				voffset = va_arg(ap, long *);
+				vbin = va_arg(ap, long *);
+
+				if (width != NULL)
+					*width = cam->image_area.lr.x - cam->image_area.ul.x;
+
+				if (hoffset != NULL)
+					*hoffset = cam->image_area.ul.x;
+
+				if (hbin != NULL)
+					*hbin = cam->hbin;
+
+				if (height != NULL)
+					*height = cam->image_area.lr.y - cam->image_area.ul.y;
+
+				if (voffset != NULL)
+					*voffset = cam->image_area.ul.y;
+
+				if (vbin != NULL)
+					*vbin = cam->vbin;
+
+				r = 0;
+			}
+			break;
+
+
 		case FLI_CONTROL_SHUTTER:
 			if (argc != 1)
 				r = -EINVAL;
@@ -724,18 +948,21 @@ long fli_camera_command(flidev_t dev, int cmd, int argc, ...)
 
 				shutter = *va_arg(ap, long *);
 
-				if( (shutter == FLI_SHUTTER_EXTERNAL_TRIGGER_LOW) ||
-					  (shutter == FLI_SHUTTER_EXTERNAL_TRIGGER_HIGH) )
+				if( (shutter & FLI_SHUTTER_EXTERNAL_TRIGGER_LOW) ||
+					  (shutter & FLI_SHUTTER_EXTERNAL_TRIGGER_HIGH) ||
+						((shutter & FLI_SHUTTER_EXTERNAL_EXPOSURE_CONTROL)) )
 				{
-					debug(FLIDEBUG_INFO, "External trigger.");
+					debug(FLIDEBUG_INFO, "External trigger: %02x", shutter);
 					cam->exttrigger = 1;
-					cam->exttriggerpol = (shutter == FLI_SHUTTER_EXTERNAL_TRIGGER_LOW)?0:1;
+					cam->exttriggerpol = (shutter & FLI_SHUTTER_EXTERNAL_TRIGGER_LOW)?0:1;
+					cam->extexposurectrl = (shutter & FLI_SHUTTER_EXTERNAL_EXPOSURE_CONTROL)?1:0; 
 					r = 0;
 				}
 				else
 				{
 					debug(FLIDEBUG_INFO, "No External trigger.\n");
 					cam->exttrigger = 0;
+					cam->extexposurectrl = 0;
 
 					switch (DEVICE->domain)
 					{
@@ -794,7 +1021,7 @@ long fli_camera_command(flidev_t dev, int cmd, int argc, ...)
 				switch (DEVICE->domain)
 				{
 					case FLIDOMAIN_PARALLEL_PORT:
-						r = -EFAULT;
+						r = -EINVAL;
 						break;
 
 					case FLIDOMAIN_USB:
@@ -923,6 +1150,110 @@ long fli_camera_command(flidev_t dev, int cmd, int argc, ...)
 			}
 			break;
 
+		case FLI_TRIGGER_EXPOSURE:
+			if (argc != 0)
+				r = -EINVAL;
+			else
+			{
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_PARALLEL_PORT:
+						r = -EINVAL;
+						break;
+
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_trigger_exposure(dev);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
+		case FLI_END_EXPOSURE:
+			if (argc != 0)
+				r = -EINVAL;
+			else
+			{
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_PARALLEL_PORT:
+						r = -EINVAL;
+						break;
+
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_end_exposure(dev);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
+		case FLI_READ_EEPROM:
+			if (argc != 4)
+				r = -EINVAL;
+			else
+			{
+				long loc;
+				long address;
+				long length;
+				void *rbuf;
+
+				loc = *va_arg(ap, long *);
+				address = *va_arg(ap, long *);
+				length = *va_arg(ap, long *);
+				rbuf = va_arg(ap, void *);
+
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_PARALLEL_PORT:
+						r = -EINVAL;
+						break;
+
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_read_eeprom(dev, loc, address, length, rbuf);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
+		case FLI_WRITE_EEPROM:
+			if (argc != 4)
+				r = -EINVAL;
+			else
+			{
+				long loc;
+				long address;
+				long length;
+				void *rbuf;
+
+				loc = *va_arg(ap, long *);
+				address = *va_arg(ap, long *);
+				length = *va_arg(ap, long *);
+				rbuf = va_arg(ap, void *);
+
+				switch (DEVICE->domain)
+				{
+					case FLIDOMAIN_PARALLEL_PORT:
+						r = -EINVAL;
+						break;
+
+					case FLIDOMAIN_USB:
+						r = fli_camera_usb_write_eeprom(dev, loc, address, length, rbuf);
+						break;
+
+					default:
+						r = -EINVAL;
+				}
+			}
+			break;
+
 		default:
 			r = -EINVAL;
   }
@@ -950,7 +1281,7 @@ static long fli_camera_set_frame_type(flidev_t dev, fliframe_t frametype)
 
   cam = DEVICE->device_data;
 
-  if ((frametype < FLI_FRAME_TYPE_NORMAL) || (frametype > FLI_FRAME_TYPE_DARK))
+  if (frametype & 0xfff8)
     return -EINVAL;
 
   cam->frametype = frametype;
